/*
 * Copyright (c) 2013, Pushkar Singh <psingh@sangoma.com>
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Contributors: Pushkar Singh <psingh@sangoma.com>
 */

#include "sangoma_gsm.h"

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
void on_sigint(int signal);
void on_sigusr(int signal);

void create_span(ftdm_span_t *span);
void post_answer_event(ftdm_channel_t *ftdmchan);

int gsm_usr_opt(void);
void send_sms(void);
void call_req(void);
int print_status(void);


/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;
span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE] = {{0}};

static ftdm_hash_t *gsm_configs;

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

/**********************************************************************************
 * Fun   : load_gsm_config()
 * Desc  : Load configuration with respect to GSM modules and start the associated
 * 	   span.
 * Ret   : GSM_SUCCESS | GSM_FAIL
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
static int load_gsm_config(void)
{
	ftdm_span_t *span = NULL;
	const char *name = "wp3";
	unsigned paramindex = 0;
	ftdm_conf_parameter_t spanparameters[30];
	ftdm_status_t zstatus = FTDM_FAIL;
	uint32_t span_id = 0;

	/* go through with avavilable physical wanpipe ports and prepare span list */
	/* TODO - name has to proper */
	/* Step - 1 :
	 * "Found 4 ports in system ... print port information ..
	 * Enter span name for port:1 : <span_name>
	 */

	if (!name) {
		ftdm_log(FTDM_LOG_ERROR, "Span missing required attibute 'span_name' invalid ...\n");
		return GSM_FAILURE;
	}

	/* span is already configured through global_configuration api from freetdm.conf */
	zstatus = ftdm_span_find_by_name(name, &span);

	if (zstatus != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error finding FreeTDM span name: %s\n", name);
		return GSM_FAILURE;
	}

	/* Get the span_id of the respective span that needs to be configured */
	span_id = ftdm_span_get_id(span);

	/* Set moduleType type */
	spanparameters[paramindex].var = "moduleType";
	spanparameters[paramindex].val = "telit";
	paramindex++;

	/* The last parameter .var member must be NULL! */
	spanparameters[paramindex].var = NULL;
	spanparameters[paramindex].val = NULL;

	/* Configure span */
	if (ftdm_configure_span_signaling(span, "gsm", on_gsm_signal, spanparameters) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Error configuring Sangoma GSM FreeTDM span %d\n", span_id);
		return GSM_FAILURE;
	}

	/* Initializing Structure variables w.r.t the span configured */
	SPANS[span_id].inuse = 1;
	SPANS[span_id].ftdmspan = span;
	ftdm_copy_string(SPANS[span_id].type, "GSM", sizeof(SPANS[span_id].type));
	ftdm_log(FTDM_LOG_DEBUG, "Configured GSM FreeTDM span %d\n", span_id);
	fprintf(stderr, "Configured GSM FreeTDM span %d\n", span_id);

	/* Start the configured span */
	ftdm_span_start(span);

	return GSM_SUCCESS;
}


/**********************************************************************************
 * Fun   : create_span()
 * Desc  : Create and configure spans depending upon the wanpipe configured
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
void create_span(ftdm_span_t *span)
{
	/* Initializing local variables */
	unsigned chan_configured = 0;
	unsigned configured = 0;
	char wpchans[25];
	ftdm_channel_config_t chan_config;

	memset(&wpchans, 0, sizeof(wpchans));
	memset(&chan_config, 0, sizeof(chan_config));

	/* Create and configure Span depedning upon the wanpipe configured */
	if (ftdm_span_create("wanpipe", "wp3", &span) != FTDM_SUCCESS) {
		fprintf(stderr, "unable to create span w3g1 of type GSM\n");
		exit(EXIT_FAILURE);
	}

	fprintf(stdout, "Successfully created span w3g1 of type GSM\n");

	/* Setting span trunk type */
	ftdm_span_set_trunk_type(span, FTDM_TRUNK_E1);

	/* Configuring span b-channel */
	chan_config.type = FTDM_CHAN_TYPE_B;
	snprintf(wpchans, sizeof(wpchans), "3:1");

	ftdm_configure_span_channels(span, "3:1", &chan_config, &chan_configured);
	fprintf(stdout, "registered << %u << B channels", chan_configured);

	configured +=chan_configured;

	/* Configuring span d-channel */
	chan_configured = 0;

	chan_config.type = FTDM_CHAN_TYPE_DQ921;
	snprintf(wpchans, sizeof(wpchans), "3:2");

	ftdm_configure_span_channels(span, "3:2", &chan_config, &chan_configured);
	fprintf(stdout, "registered << %u << D channels", chan_configured);

	configured +=chan_configured;

	fprintf(stdout, "Configured %u channels ...\n", configured);
	if (!configured) {
		exit(EXIT_FAILURE);
	}
}


/**********************************************************************************
 * Fun   : post_answer_event()
 * Desc  : Answer the incoming call
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
void post_answer_event(ftdm_channel_t *ftdmchan)
{
	ftdm_set_state_locked(ftdmchan, FTDM_CHANNEL_STATE_UP);
}

/* TODO: Set state to Down if required */
/*void post_hangup_event(ftdm_channel_t *ftdmchan)
{
	ftdm_set_state_locked(ftdmchan, FTDM_CHANNEL_STATE_DOWN);
}*/


/**********************************************************************************
 * Fun   : on_gsm_signal()
 * Desc  : Handle events liks receive SMS, Calls etc. and display information 
 *	   accordingly
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
FIO_SIGNAL_CB_FUNCTION(on_gsm_signal)
{
	unsigned int span_id, chan_id;
	ftdm_channel_t *ftdmchan = sigmsg->channel;

	/* Get the resoected span_id and channel_id */
	span_id = ftdm_channel_get_ph_span_id(ftdmchan);
        chan_id = ftdm_channel_get_ph_id(ftdmchan);

	/* Handling different events received */
	switch (sigmsg->event_id) {
		case FTDM_SIGEVENT_SMS:
			{
				ftdm_caller_data_t *caller_data = ftdm_channel_get_caller_data(sigmsg->channel);
				ftdm_sms_data_t *sms = (ftdm_sms_data_t*) caller_data->priv;

				ftdm_log(FTDM_LOG_INFO, "SMS Recieved from %s number: %s", sms->from, sms->body);
				break;
			}

		case FTDM_SIGEVENT_START:
			{
				ftdm_caller_data_t *caller_data = ftdm_channel_get_caller_data(sigmsg->channel);
				ftdm_log(FTDM_LOG_INFO, "%s Calling\n", caller_data->dnis.digits);
				post_answer_event(ftdmchan);
				break;
			}

		case FTDM_SIGEVENT_STOP:
			{
				ftdm_log(FTDM_LOG_INFO, "Call hangup event received \n");
				break;
			}

		case FTDM_SIGEVENT_RELEASED:
			{
				ftdm_log(FTDM_LOG_INFO, "Call Released event received \n");
				break;
			}

		case FTDM_SIGEVENT_ALARM_CLEAR:
			{
				ftdm_log(FTDM_LOG_INFO,"Received FTDM_SIGEVENT_ALARM_CLEAR \n");
				break;
			}

		case FTDM_SIGEVENT_ALARM_TRAP:
			{
				ftdm_alarm_flag_t alarmbits = FTDM_ALARM_NONE;
				char alarm_type[80];
				if (ftdm_channel_get_alarms(sigmsg->channel, &alarmbits) != FTDM_SUCCESS) {
					ftdm_log(FTDM_LOG_ERROR, "FTDM_SIGEVENT_ALARM_TRAP: failed to retrieve alarms\n");
					return FTDM_FAIL;
				}

				strcpy(alarm_type,"UNKNOWN");

				if (alarmbits & FTDM_ALARM_RED) {
					strcpy(alarm_type,"RED");
				}

				if (alarmbits & FTDM_ALARM_YELLOW) {
					strcpy(alarm_type,"YELLOW");
				}

				if (alarmbits & FTDM_ALARM_RAI) {
					strcpy(alarm_type,"RAI");
				}

				if (alarmbits & FTDM_ALARM_BLUE) {
					strcpy(alarm_type,"BLUE");
				}

				if (alarmbits & FTDM_ALARM_AIS) {
					strcpy(alarm_type,"AIS");
				}

				if (alarmbits & FTDM_ALARM_GENERAL) {
					strcpy(alarm_type,"GENERAL");
				}

				ftdm_log(FTDM_LOG_INFO,"Received FTDM_SIGEVENT_ALARM_TRAP with Alarm=%s\n", alarm_type);
				break;
			}

		case FTDM_SIGEVENT_DIALING:
			ftdm_log(FTDM_LOG_INFO,"Received FTDM_SIGEVENT_DIALING in state DIALING\n");
			break;

		case FTDM_SIGEVENT_RINGING:
			ftdm_log(FTDM_LOG_INFO,"Received FTDM_SIGEVENT_RINGING in state RINGING\n");
			break;
		default:
			ftdm_log(FTDM_LOG_INFO,"Invalid Event Id received = %d\n", sigmsg->event_id);
			break;
	}

	return FTDM_SUCCESS;
}


/**********************************************************************************
 * Fun   : on_sigusr()
 * Desc  : Handling signals due to user interruption
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
void on_sigusr(int signal)
{
	switch(signal) {
		case SIGUSR1:
			fprintf(stdout, "User interrupted... \n");
		break;

		case SIGUSR2:
			fprintf(stderr, "User interrupted... \n");
		break;
	}

	fprintf(stdout, "SANGOMA GSM MODULE Application is Getting closed\n");
	ftdm_global_destroy();
	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");
	getchar();
	fprintf(stdout, "Press enter... \n");
	exit(EXIT_FAILURE);
}


/**********************************************************************************
 * Fun   : send_sms()
 * Desc  : For sending SMS
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
void send_sms(void)
{
	char *rply = ftdm_api_execute("gsm sms wp3 8867293028 Hello Sir");
	if (rply) {
		fprintf(stdout, "%s", rply);
		ftdm_free(rply);
	}

	ftdm_sleep(4000);
}


/**********************************************************************************
 * Fun   : call_req()
 * Desc  : For Calling to a specific number
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
void call_req(void)
{
	char *rply = ftdm_api_execute("gsm con 1 8867293028 7848035573");
	if (rply) {
		fprintf(stdout, "%s", rply);
		ftdm_free(rply);
	}

	ftdm_sleep(4000);
}


/**********************************************************************************
 * Fun   : main()
 * Desc  : Triggers functions in order to do the Wanpipe configurationa and freeTDM
 * 	   configuration and provide a UI environment to user in order to perform
 * 	   various functions w.r.t. GSM.
 * Ret   : GSM_SUCCESS | GSM_FAIL
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
int main(void)
{
	/* local variables initialization */
	int max_wait = 20;
	ftdm_span_t *span = NULL;

	memset(SPANS, 0, FTDM_MAX_SPANS_INTERFACE*(sizeof(span_data_t)));

	/* Signal handler in order to handle different signal */
	signal(SIGINT, on_sigint);
	signal(SIGUSR1, on_sigusr);
	signal(SIGUSR2, on_sigusr);

	/* ftdm_global_set_default_logger(FTDM_LOG_LEVEL_DEBUG); */
	ftdm_global_set_default_logger(FTDM_LOG_LEVEL_INFO);
	ftdm_global_set_crash_policy(FTDM_CRASH_ON_ASSERT);

	/* TODO : Step-1 : Use libsangoma API to discover and start the GSM ports (using a pseudo-API we create in another file to keep this file simple)
	 */

	/* Now you have list of GSM Ports */

	gsm_configs = create_hashtable(16, ftdm_hash_hashfromstring, ftdm_hash_equalkeys);

	/* Freetdm global initialization */
	if (ftdm_global_init() != FTDM_SUCCESS) {
		fprintf(stderr, "Error loading FreeTDM\n");
		exit(EXIT_FAILURE);
	}

	/* Create spans in FreeTDM */
	create_span(span);

	/* Loads the configuration with respect to GSM module */
	if (load_gsm_config()) {
		fprintf(stderr, "Failed to load configurations\n");
		ftdm_global_destroy();
		return GSM_FAILURE;
	}

	ftdm_sleep(3000);
	fprintf(stdout, "GSM module started ...\n");

	/* Provide diferent options to user to perform various operation depending up on the option
	 * selected by user */
	while (gsm_usr_opt()) {
		continue;
	}

	while (g_running) {
		ftdm_sleep(1000);
	}

	while (g_num_threads && (max_wait-- > 0)) {
		fprintf(stderr, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}


	ftdm_global_destroy();
	return GSM_SUCCESS;
}


/**********************************************************************************
 * Fun   : on_sigint()
 * Desc  : Handling SIGINT event
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
void on_sigint(int signal)
{
	g_running = 0;

	fprintf(stdout, "SIGINT recieved ...\n");
	fprintf(stdout, "SANGOMA GSM MODULE Application is Getting closed\n");
	ftdm_global_destroy();
	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                      GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Press Enter key to exit... \n");
	getchar();
	exit(EXIT_FAILURE);
}


/**********************************************************************************
 * Fun   : gsm_usr_opt()
 * Desc  : Provide various options to perform different operations as per user's
 * 	   choice.
 * Ret   : GSM_FAILURE
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
int gsm_usr_opt(void)
{
	char choice[6];

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "		WELCOME TO SANGOMA GSM MODULE             \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Please select any one of the below options :\n");

	fprintf(stdout, "1. Send SMS\n2. Call To A Number\n3. Print Span Information\n4. Exit\n");
	fprintf(stdout, "Enter your choice[1-4]: ");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			fprintf(stderr, "Invalid Input Choice must must be [1-4]...\n");
			return GSM_FAILURE;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Send SMS
		 * 2) Print Span related information
		 * 3) Exit from application
		 */
		switch (choice[0]) {
			case '1':
				send_sms();
				break;

			case '2':
				call_req();
				break;

			case '3':
				while (print_status()) {
					continue;
				}
				break;

			case '4':
				fprintf(stdout, "GSM Sample Application is getting closed...\n");
				ftdm_global_destroy();
				fprintf(stdout, "\n\n\n=========================================================\n");
				fprintf(stdout, "                      GOOD BYE                            \n");
				fprintf(stdout, "=========================================================\n");
				exit(EXIT_FAILURE);
				break;

			default:
				fprintf(stderr, "Invalid Option. Please Choose a correct option value...\n");
				break;
		}
	}

	return GSM_FAILURE;
}


/**********************************************************************************
 * Fun   : print_status()
 * Desc  : Prints ftdm span and gsm span status
 * Ret   : void
 * Note  :
 * Author: Pushkar Singh
 **********************************************************************************/
int print_status(void)
{
	char choice[6];
	int ret = 1;

	fprintf(stdout, "\n\n\nPlease select any one of the below options :\n");

	fprintf(stdout, "1. Physical Span list\n2. Physical Span Status\n3. GSM Span status\n4. Go Back To Previous Menu\n");
	fprintf(stdout, "Enter your choice[1-4]: ");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			fprintf(stderr, "Invalid Input Choice must must be [1-4]...\n");
			return GSM_FAILURE;
		}

		choice[1] = '\0';

		/* Provides information w.r.t spans configured dependinh upon the user request like:
		 * 1) List of Span configured with information about the respective span
		 * 2) Information with respect to a particular span as requested by user
		 * 3) Infotmation with respect to GSM span as requested by user
		 * 4) Go back to Previous Meanu
		 */
		switch (choice[0]) {
			case '1':
				{
					ret = ftdm_list();
					ftdm_sleep(100);
				}
				break;

			case '2':
				{
					ret = ftdm_dump();
					ftdm_sleep(100);
				}
				break;

			case '3':
				{
					ret = gsm_span_status();
					ftdm_sleep(100);
				}
				break;

			case '4':
				return GSM_SUCCESS;
				break;

			default:
				fprintf(stderr, "Invalid Option. Please Choose a correct option value...\n");
				break;
		}
	}

	return ret;
}
